### System ###
import csv
from io import StringIO, BytesIO

### Local ###
from .midi.events import *
from .midi.containers import *
from .events import csv_to_midi_map
from .midi.fileio import FileWriter

COMMENT_DELIMITERS = ("#", ";")


def parse(file):
    """
    Input is a string, file or file-like object of type stringIO or bytesIO.
    Output is a file-like object of type bytesIO containing the binary MIDI data.
    """
    pattern = Pattern(tick_relative=False)
    for line in csv.reader(file, skipinitialspace=True):
        if line[0].startswith(COMMENT_DELIMITERS):
            continue
        tr = int(line[0])
        time = int(line[1])
        identifier = line[2].strip()
        if identifier == "Header":
            pattern.resolution = int(line[5])
        elif identifier == "End_of_file":
            continue
        elif identifier == "Start_track":
            track = Track(tick_relative=False)
            pattern.append(track)
        else:
            event = csv_to_midi_map[identifier](tr, time, identifier, line[3:])
            track.append(event)
    pattern.make_ticks_rel()
    return pattern


def main(args):
    with open(args.input, "r") as input_file:
        pattern = parse(input_file)
        with open(args.output, "wb") as output_file:
            midi_writer = FileWriter(output_file)
            midi_writer.write(pattern)


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description="Converts a CSV file into MIDI representation.")
    parser.add_argument("-i", "--input", type=str, dest="input", required=True,
                        metavar="file", help="(required) The file to convert to MIDI")
    parser.add_argument("-o", "--output", type=str, dest="output", required=True,
                        metavar="file", help="(required) The file to write the output to")
    args = parser.parse_args()

    main(args)



def read_varlen(data):
    NEXTBYTE = 1
    value = 0
    while NEXTBYTE:
        chr = next(data)
        # is the hi-bit set?
        if not (chr & 0x80):
            # no next BYTE
            NEXTBYTE = 0
        # mask out the 8th bit
        chr = chr & 0x7f
        # shift last value up 7 bits
        value = value << 7
        # add new value
        value += chr
    return value


def write_varlen(value):
    b1 = value & 0x7F
    value >>= 7
    if value:
        b2 = (value & 0x7F) | 0x80
        value >>= 7
        if value:
            b3 = (value & 0x7F) | 0x80
            value >>= 7
            if value:
                b4 = (value & 0x7F) | 0x80
                res = bytes((b4, b3, b2, b1))
            else:
                res = bytes((b3, b2, b1))
        else:
            res = bytes((b2, b1))
    else:
        res = bytes((b1,))
    return res



from .midi.events import *
from .csv_converters import *
from .midi_converters import *

midi_to_csv_map = {
    NoteOffEvent: from_NoteOffEvent,
    NoteOnEvent: from_NoteOnEvent,
    AfterTouchEvent: from_AfterTouchEvent,
    ControlChangeEvent: from_ControlChangeEvent,
    ProgramChangeEvent: from_ProgramChangeEvent,
    ChannelAfterTouchEvent: from_ChannelAfterTouchEvent,
    PitchWheelEvent: from_PitchWheelEvent,
    SequenceNumberMetaEvent: from_SequenceNumberMetaEvent,
    ProgramNameEvent: from_ProgramNameEvent,
    TextMetaEvent: from_TextMetaEvent,
    CopyrightMetaEvent: from_CopyrightMetaEvent,
    TrackNameEvent: from_TrackNameEvent,
    InstrumentNameEvent: from_InstrumentNameEvent,
    LyricsEvent: from_LyricsEvent,
    MarkerEvent: from_MarkerEvent,
    CuePointEvent: from_CuePointEvent,
    ChannelPrefixEvent: from_ChannelPrefixEvent,
    PortEvent: from_PortEvent,
    EndOfTrackEvent: from_EndOfTrackEvent,
    SomethingEvent: from_SomethingEvent,
    TrackLoopEvent: from_TrackLoopEvent,
    SetTempoEvent: from_SetTempoEvent,
    SmpteOffsetEvent: from_SmpteOffsetEvent,
    TimeSignatureEvent: from_TimeSignatureEvent,
    KeySignatureEvent: from_KeySignatureEvent,
    SequencerSpecificEvent: from_SequencerSpecificEvent,
    SysexEvent: from_SysexEvent,
}

csv_to_midi_map = {
    "Note_off_c": to_NoteOffEvent,
    "Note_on_c": to_NoteOnEvent,
    "Poly_aftertouch_c": to_AfterTouchEvent,
    "Control_c": to_ControlChangeEvent,
    "Program_c": to_ProgramChangeEvent,
    "Channel_aftertouch_c": to_ChannelAfterTouchEvent,
    "Pitch_bend_c": to_PitchWheelEvent,
    "Sequence_number": to_SequenceNumberMetaEvent,
    "Program_name_t": to_ProgramNameEvent,
    "Text_t": to_TextMetaEvent,
    "Copyright_t": to_CopyrightMetaEvent,
    "Title_t": to_TrackNameEvent,
    "Instrument_name_t": to_InstrumentNameEvent,
    "Lyric_t": to_LyricsEvent,
    "Marker_t": to_MarkerEvent,
    "Cue_point_t": to_CuePointEvent,
    "Channel_prefix": to_ChannelPrefixEvent,
    "MIDI_port": to_PortEvent,
    "End_track": to_EndOfTrackEvent,
    "Something": to_SomethingEvent,
    "Loop_track": to_TrackLoopEvent,
    "Tempo": to_SetTempoEvent,
    "SMPTE_offset": to_SmpteOffsetEvent,
    "Time_signature": to_TimeSignatureEvent,
    "Key_signature": to_KeySignatureEvent,
    "Sequencer_specific": to_SequencerSpecificEvent,
    "System_exclusive": to_SysexEvent,
}



### System ###
from io import StringIO

### Local ###
from .events import midi_to_csv_map
from .midi.fileio import read_midifile


def parse(file):
    """
    Input is a string, file or file-like object of type stringIO or bytesIO.
    Output is a file-like object of type stringIO containing the CSV data.
    """
    csv_file = StringIO()
    pattern = read_midifile(file)
    csv_file.write("0, 0, Header, {}, {}, {}\n".format(pattern.format, len(pattern), pattern.resolution))
    for index, track in enumerate(pattern):
        csv_file.write("{}, {}, Start_track\n".format(index + 1, 0))
        abstime = 0
        for event in track:
            abstime += event.tick
            csv_file.write(midi_to_csv_map[type(event)](index + 1, abstime, event))
    csv_file.write("0, 0, End_of_file")
    return csv_file


def main(args):
    with open(args.input, "rb") as input_file:
        converted = parse(input_file)
        with open(args.output, "w") as output_file:
            output_file.write(converted.getvalue())


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description="Converts a MIDI file into a CSV representation.")
    parser.add_argument("-i", "--input", type=str, dest="input", required=True,
                        metavar="file", help="(required) The file to convert to CSV")
    parser.add_argument("-o", "--output", type=str, dest="output", required=True,
                        metavar="file", help="(required) The file to write the output to")
    args = parser.parse_args()

    main(args)



from .containers import *
from .events import *
from struct import unpack, pack
from .constants import *
from .util import *


class FileReader(object):

    def read(self, midifile):
        pattern = self.parse_file_header(midifile)
        for track in pattern:
            self.parse_track(midifile, track)
        return pattern

    def parse_file_header(self, midifile):
        # First four bytes are MIDI header
        magic = midifile.read(4)
        if magic != b'MThd':
            raise TypeError("Bad header in MIDI file.", magic)
        # next four bytes are header size
        # next two bytes specify the format version
        # next two bytes specify the number of tracks
        # next two bytes specify the resolution/PPQ/Parts Per Quarter
        # (in other words, how many ticks per quater note)
        data = unpack(">LHHH", midifile.read(10))
        hdrsz = data[0]
        format = data[1]
        tracks = [Track() for x in range(data[2])]
        resolution = data[3]
        # XXX: the assumption is that any remaining bytes
        # in the header are padding
        if hdrsz > DEFAULT_MIDI_HEADER_SIZE:
            midifile.read(hdrsz - DEFAULT_MIDI_HEADER_SIZE)
        return Pattern(tracks=tracks, resolution=resolution, format=format)

    def parse_track_header(self, midifile):
        # First four bytes are Track header
        magic = midifile.read(4)
        if magic != b'MTrk':
            raise TypeError("Bad track header in MIDI file: ", magic)
        # next four bytes are track size
        trksz = unpack(">L", midifile.read(4))[0]
        return trksz

    def parse_track(self, midifile, track):
        self.RunningStatus = None
        trksz = self.parse_track_header(midifile)
        trackdata = iter(midifile.read(trksz))
        while True:
            try:
                event = self.parse_midi_event(trackdata)
                track.append(event)
            except StopIteration:
                break

    def parse_midi_event(self, trackdata):
        # first datum is varlen representing delta-time
        tick = read_varlen(trackdata)
        # next byte is status message
        stsmsg = next(trackdata)
        # is the event a MetaEvent?
        if MetaEvent.is_event(stsmsg):
            cmd = next(trackdata)
            if cmd not in EventRegistry.MetaEvents:
                raise Warning("Unknown Meta MIDI Event: " + repr(cmd))
            cls = EventRegistry.MetaEvents[cmd]
            datalen = read_varlen(trackdata)
            data = [next(trackdata) for x in range(datalen)]
            return cls(tick=tick, data=data)
        # is this event a Sysex Event?
        elif SysexEvent.is_event(stsmsg):
            data = []
            while True:
                datum = next(trackdata)
                if datum == 0xF7:
                    break
                data.append(datum)
            return SysexEvent(tick=tick, data=data)
        # not a Meta MIDI event or a Sysex event, must be a general message
        else:
            key = stsmsg & 0xF0
            if key not in EventRegistry.Events:
                assert self.RunningStatus, ("Bad byte value", tick, stsmsg, bytes(trackdata))
                data = []
                key = self.RunningStatus & 0xF0
                cls = EventRegistry.Events[key]
                channel = self.RunningStatus & 0x0F
                data.append(stsmsg)
                data += [next(trackdata) for x in range(cls.length - 1)]
                return cls(tick=tick, channel=channel, data=data)
            else:
                self.RunningStatus = stsmsg
                cls = EventRegistry.Events[key]
                channel = self.RunningStatus & 0x0F
                data = [next(trackdata) for x in range(cls.length)]
                return cls(tick=tick, channel=channel, data=data)
        raise Warning("Unknown MIDI Event: " + repr(stsmsg))


class FileWriter(object):
    RunningStatus = None

    def __init__(self, file):
        self.file = file

    def write(self, pattern):
        self.write_file_header(pattern, len(pattern))
        for track in pattern:
            self.write_track(track)

    def write_file_header(self, pattern, length=None):
        if length is None:
            length = len(pattern)
        # First four bytes are MIDI header
        packdata = pack(">LHHH", 6,
                        pattern.format,
                        length,
                        pattern.resolution)
        self.file.write(b'MThd' + packdata)

    def write_track(self, track):
        hlen = len(self.encode_track_header(0))
        buf = bytearray(b'0' * hlen)
        for event in track:
            buf.extend(self.encode_midi_event(event))
        buf[:hlen] = self.encode_track_header(len(buf) - hlen)
        self.file.write(buf)

    def write_track_header(self, track=None):
        trklen = 1 if track is None else track if isinstance(track, int) else len(track)
        self.file.write(self.encode_track_header(trklen))

    def encode_track_header(self, trklen):
        return b'MTrk' + pack(">L", trklen)

    def write_midi_event(self, event):
        # be sure to write the track and pattern headers first
        # can stream to timidity or fluidsynth this way
        self.file.write(self.encode_midi_event(event))

    def encode_midi_event(self, event):
        ret = bytearray()
        #assert hasattr(event,'tick'), event
        assert isinstance(event.tick, int), event.tick
        ret.extend(write_varlen(event.tick))
        # is the event a MetaEvent?
        if isinstance(event, MetaEvent):
            ret.append(event.statusmsg)
            ret.append(event.metacommand)
            ret.extend(write_varlen(len(event.data)))
            ret.extend(event.data)
        # is this event a Sysex Event?
        elif isinstance(event, SysexEvent):
            ret.append(0xF0)
            ret.extend(event.data)
            ret.append(0xF7)
        # not a Meta MIDI event or a Sysex event, must be a general message
        elif isinstance(event, Event):
            # why in the heeeeeeeeelp would you not write the status message
            # here? doesn't matter if it's the same as last time. the byte
            # needs to be there!
            ret.append(event.statusmsg | event.channel)
            ret.extend(event.data)
        else:
            raise ValueError("Unknown MIDI Event: " + str(event))
        return ret


def write_midifile(midifile, pattern):
    if type(midifile) in (str, str):
        with open(midifile, 'wb') as out:
            return write_midifile(out, pattern)
    writer = FileWriter(midifile)
    return writer.write(pattern)


def read_midifile(midifile):
    if type(midifile) in (str, bytes):
        with open(midifile, 'rb') as inp:
            return read_midifile(inp)
    reader = FileReader()
    return reader.read(midifile)



### Local ###
from .midi.events import *


def text_encode(text):
    encoded = ""
    for character in text:
        if (ord(character) < 32) or (ord(character) > 128):
            for byte in character.encode("utf8"):
                encoded += "\\{:03o}".format(byte)
        else:
            encoded += character
    return encoded


def write_event(track, time, identifier, data):
    return ("{}, {}, {}" + (", {}" * len(data)) + "\n").format(track, time, identifier, *data)


def from_NoteOffEvent(track, time, event):
    return write_event(track, time, "Note_off_c", [event.channel, *event.data])


def from_NoteOnEvent(track, time, event):
    return write_event(track, time, "Note_on_c", [event.channel, *event.data])


def from_AfterTouchEvent(track, time, event):
    return write_event(track, time, "Poly_aftertouch_c", [event.channel, *event.data])


def from_ControlChangeEvent(track, time, event):
    return write_event(track, time, "Control_c", [event.channel, *event.data])


def from_ProgramChangeEvent(track, time, event):
    return write_event(track, time, "Program_c", [event.channel, *event.data])


def from_ChannelAfterTouchEvent(track, time, event):
    return write_event(track, time, "Channel_aftertouch_c", [event.channel, event.data[0]])


def from_PitchWheelEvent(track, time, event):
    return write_event(track, time, "Pitch_bend_c", [event.channel, (event.data[0] | (event.data[1] << 7))])


def from_SequenceNumberMetaEvent(track, time, event):
    return write_event(track, time, "Sequence_number", [((event.data[0] << 8) | event.data[1])])


def from_ProgramNameEvent(track, time, event):
    return write_event(track, time, "Program_name_t", ['"{}"'.format(text_encode(event.text))])


def from_TextMetaEvent(track, time, event):
    return write_event(track, time, "Text_t", ['"{}"'.format(text_encode(event.text))])


def from_CopyrightMetaEvent(track, time, event):
    return write_event(track, time, "Copyright_t", ['"{}"'.format(text_encode(event.text))])


def from_TrackNameEvent(track, time, event):
    return write_event(track, time, "Title_t", ['"{}"'.format(text_encode(event.text))])


def from_InstrumentNameEvent(track, time, event):
    return write_event(track, time, "Instrument_name_t", ['"{}"'.format(text_encode(event.text))])


def from_LyricsEvent(track, time, event):
    return write_event(track, time, "Lyric_t", ['"{}"'.format(text_encode(event.text))])


def from_MarkerEvent(track, time, event):
    return write_event(track, time, "Marker_t", ['"{}"'.format(text_encode(event.text))])


def from_CuePointEvent(track, time, event):
    return write_event(track, time, "Cue_point_t", ['"{}"'.format(text_encode(event.text))])


def from_ChannelPrefixEvent(track, time, event):
    return write_event(track, time, "Channel_prefix", [*event.data])


def from_PortEvent(track, time, event):
    return write_event(track, time, "MIDI_port", [*event.data] if event.data else [0])


def from_EndOfTrackEvent(track, time, event):
    return write_event(track, time, "End_track", [])


def from_SomethingEvent(track, time, event):
    return write_event(track, time, "Something", [])


def from_TrackLoopEvent(track, time, event):
    return write_event(track, time, "Loop_track", [])


def from_SetTempoEvent(track, time, event):
    return write_event(track, time, "Tempo", [event.get_mpqn()])


def from_SmpteOffsetEvent(track, time, event):
    return write_event(track, time, "SMPTE_offset", [*event.data])


def from_TimeSignatureEvent(track, time, event):
    return write_event(track, time, "Time_signature", [*event.data])


def from_KeySignatureEvent(track, time, event):
    return write_event(track, time, "Key_signature", [event.get_alternatives(), '"major"' if event.data[1] == 0 else '"minor"'])


def from_SequencerSpecificEvent(track, time, event):
    return write_event(track, time, "Sequencer_specific", [len(event.data), *event.data])


def from_SysexEvent(track, time, event):
    return write_event(track, time, "System_exclusive", [len(event.data), *event.data])



from setuptools import setup

setup(name='py_midicsv',
    version='1.7.0',
    description='A library for converting MIDI files from and to CSV format.',
    long_description='''
        py_midicsv is a Python port of the tools midicsv and csvmidi (http://www.fourmilab.ch/webtools/midicsv/),
        which enable the conversion of MIDI files from and to CSV files, making them easier to work with.
        The main advantage to representing MIDI events as text is the easy application
        of text processing to these events, making MIDI files easy to work with and quick to modify.
        ''',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Multimedia :: Sound/Audio :: MIDI',
        ],
    keywords='midi converter csv',
    url='https://github.com/timwedde/py_midicsv',
    author='Tim Wedde',
    author_email='timwedde@icloud.com',
    license='MIT',
    packages=['py_midicsv', 'py_midicsv.midi'],
    # install_requires=[],
    include_package_data=True,
    zip_safe=False
)



### System ###
import re

### Local ###
from .midi.events import *


def text_decode(text):
    decoded = text
    for octc in re.findall(r"\\(\d{3})", decoded):
        decoded = decoded.replace(r"\%s" % octc, chr(int(octc, 8)))
    return decoded


def to_NoteOffEvent(track, time, identifier, line):
    channel, pitch, velocity = map(int, line)
    return NoteOffEvent(tick=time, channel=channel, pitch=pitch, velocity=velocity)


def to_NoteOnEvent(track, time, identifier, line):
    channel, pitch, velocity = map(int, line)
    return NoteOnEvent(tick=time, channel=channel, pitch=pitch, velocity=velocity)


def to_AfterTouchEvent(track, time, identifier, line):
    cannel, value = map(int, line)
    return AfterTouchEvent(tick=time, channel=channel, value=value)


def to_ControlChangeEvent(track, time, identifier, line):
    channel, control, value = map(int, line)
    return ControlChangeEvent(tick=time, channel=channel, control=control, value=value)


def to_ProgramChangeEvent(track, time, identifier, line):
    channel, value = map(int, line)
    return ProgramChangeEvent(tick=time, channel=channel, value=value)


def to_ChannelAfterTouchEvent(track, time, identifier, line):
    channel, value = map(int, line)
    return ChannelAfterTouchEvent(tick=time, channel=channel, value=value)


def to_PitchWheelEvent(track, time, identifier, line):
    channel, value = map(int, line)
    return PitchWheelEvent(tick=time, channel=channel, pitch=value - 0x2000)


def to_SequenceNumberMetaEvent(track, time, identifier, line):
    value = int(line[0])
    return SequenceNumberMetaEvent(tick=time, value=value)


def to_ProgramNameEvent(track, time, identifier, line):
    text = text_decode(line[0]).encode()
    return ProgramNameEvent(tick=time, data=text)


def to_TextMetaEvent(track, time, identifier, line):
    text = text_decode(line[0]).encode()
    return TextMetaEvent(tick=time, data=text)


def to_CopyrightMetaEvent(track, time, identifier, line):
    text = text_decode(line[0]).encode()
    return CopyrightMetaEvent(tick=time, data=text)


def to_TrackNameEvent(track, time, identifier, line):
    text = text_decode(line[0]).encode()
    return TrackNameEvent(tick=time, data=text)


def to_InstrumentNameEvent(track, time, identifier, line):
    text = text_decode(line[0]).encode()
    return InstrumentNameEvent(tick=time, data=text)


def to_LyricsEvent(track, time, identifier, line):
    text = text_decode(line[0]).encode()
    return LyricsEvent(tick=time, data=text)


def to_MarkerEvent(track, time, identifier, line):
    text = text_decode(line[0]).encode()
    return MarkerEvent(tick=time, data=text)


def to_CuePointEvent(track, time, identifier, line):
    text = text_decode(line[0]).encode()
    return CuePointEvent(tick=time, data=text)


def to_ChannelPrefixEvent(track, time, identifier, line):
    text = text_decode(line[0]).encode()
    return ChannelPrefixEvent(tick=time, data=text)


def to_PortEvent(track, time, identifier, line):
    text = text_decode(line[0]).encode()
    return PortEvent(tick=time, text=text)


def to_EndOfTrackEvent(track, time, identifier, line):
    return EndOfTrackEvent(tick=time)


def to_SomethingEvent(track, time, identifier, line):
    return SomethingEvent(tick=time)


def to_TrackLoopEvent(track, time, identifier, line):
    return TrackLoopEvent(tick=time)


def to_SetTempoEvent(track, time, identifier, line):
    mpqn = int(line[0])
    return SetTempoEvent(tick=time, mpqn=mpqn)


def to_SmpteOffsetEvent(track, time, identifier, line):
    return SmpteOffsetEvent(tick=time)


def to_TimeSignatureEvent(track, time, identifier, line):
    num, denom, click, notesq = map(int, line)
    return TimeSignatureEvent(tick=time, numerator=num, denominator=denom, metronome=click, thirtyseconds=notesq)


def to_KeySignatureEvent(track, time, identifier, line):
    key, major = int(line[0]), False if line[1] == "major" else True
    return KeySignatureEvent(tick=time, alternatives=key, minor=major)


def to_SequencerSpecificEvent(track, time, identifier, line):
    length, text = int(line[0]), text_decode(line[1]).encode()
    return SequencerSpecificEvent(tick=time, data=text)


def to_SysexEvent(track, time, identifier, line):
    length, data = int(line[0]), [int(item) for item in line[1:]]
    return SysexEvent(tick=time, data=data)
